
#include <cstdio>
#include <string>
#include <tuple>
#include <iostream>
#include "gauss.h"
#include "solver.h"


/// <summary>
/// The purpose of this template class is to invoke a given method on
/// an solver instance placed in the tuple. The ItemToCall argument identifies
/// the instance (the first instance is numbered ItemToCall, the second one ItemToCall -1 etc.).
/// 
/// This template class allows a sort of for-each over solver instances stored in the tuple
/// given as an argument.
/// </summary>
template <std::size_t ItemsToCall, typename Tuple>
class InternalSolverInvoker {
public:
	static void clear(Tuple & aTuple)
	{
		std::get<std::tuple_size<Tuple>::value - ItemsToCall>(aTuple).clear();
		InternalSolverInvoker<ItemsToCall - 1, Tuple>::clear(aTuple);
	}
	static void prepare(Tuple & aTuple)
	{
		std::get<std::tuple_size<Tuple>::value - ItemsToCall>(aTuple).prepare();
		InternalSolverInvoker<ItemsToCall - 1, Tuple>::prepare(aTuple);
	}
	template <typename NumberType>
	static void addNumber(Tuple & aTuple, NumberType aNumber)
	{
		std::get<std::tuple_size<Tuple>::value - ItemsToCall>(aTuple).addNumber(aNumber);
		InternalSolverInvoker<ItemsToCall - 1, Tuple>::addNumber<NumberType>(aTuple, aNumber);
	}
	template <typename NumberType>
	static bool compute(Tuple & aTuple, NumberType aNumbers[], std::size_t aCount)
	{
		if (std::get<std::tuple_size<Tuple>::value - ItemsToCall>(aTuple).check()) {
			std::size_t count = 0;
			std::get<std::tuple_size<Tuple>::value - ItemsToCall>(aTuple).computeNext(aCount);
			
			count = std::get<std::tuple_size<Tuple>::value - ItemsToCall>(aTuple).size();
			for (std::size_t i = 0; i < aCount; ++i)
				aNumbers[i] = std::get<std::tuple_size<Tuple>::value - ItemsToCall>(aTuple).getNumber(count - aCount + i);
		
			return true;
		}

		return InternalSolverInvoker<ItemsToCall - 1, Tuple>::compute<NumberType>(aTuple, aNumbers, aCount);
	}
};


/// <summary>
/// A stopper for the general InternalSolverInvoker.
/// </summary>
template <typename Tuple>
class InternalSolverInvoker<0, Tuple> {
public:
	static void clear(Tuple & aTuple)
	{
		return;
	}
	static void prepare(Tuple & aTuple)
	{
		return;
	}
	template <typename NumberType>
	static void addNumber(Tuple & aTuple, NumberType aNumber)
	{
		return;
	}
	template <typename NumberType>
	static bool compute(Tuple & aTuple, NumberType aNumbers[], std::size_t aCount)
	{
		return false;
	}
};


/// <summary>
/// Wraps the InternalSolverInvoker template class, so for-each iterations
/// over a tuple are quite comfortable.
/// </summary>
template <typename Tuple>
class SolverInvoker {
public:
	/// <summary>
	/// Resets all solvers in the given tuple.
	/// </summary>
	/// <param name="aTuple"></param>
	static void clear(Tuple & aTuple)
	{
		InternalSolverInvoker<std::tuple_size<Tuple>::value, Tuple>::clear(aTuple);

		return;
	}
	/// <summary>
	/// Calls the prepare() method on all solvers in the tuple.
	/// </summary>
	/// <param name="aTuple"></param>
	static void prepare(Tuple & aTuple)
	{
		InternalSolverInvoker<std::tuple_size<Tuple>::value, Tuple>::prepare(aTuple);

		return;
	}
	/// <summary>
	/// Adds an element of a sequence to all solvers.
	/// </summary>
	template <typename NumberType>
	static void addNumber(Tuple & aTuple, NumberType aNumber)
	{
		InternalSolverInvoker<std::tuple_size<Tuple>::value, Tuple>::addNumber<NumberType>(aTuple, aNumber);

		return;
	}
	/// <summary>
	/// Goes through the solvers in the tuple and finds the first of them that
	/// is capable of computing additional elements of the sequence (its check()
	/// method returned true). If such a solver exists, its computeNext() method
	/// is used to predict the sequence and retrieve the newly computed elements.
	/// </summary>
	template <typename NumberType>
	static bool compute(Tuple & aTuple, NumberType aNumbers[], std::size_t aCount)
	{
		return InternalSolverInvoker<std::tuple_size<Tuple>::value, Tuple>::compute<NumberType>(aTuple, aNumbers, aCount);
	}
};



int main(int argc, char **argv)
{
	bool done = false;
	std::string line;
	std::tuple<
		// Each element depends linearly on at most three of its predecessors
		CSolverWrapper<long long, CLinearSolver<long long, 3>>, 
		// Each element depends polynomially (max. degree is 3) on its predecessor.
		CSolverWrapper<long long, CPolynomialSolver<long long, 3>>,
		// Each element is a polynomial function of its index (max. degree is 5).
		CSolverWrapper<long long, CIndexSolver<long long, 5>>,
		// Each element depends on a combination of its index and its predecessor (at most quadratically)
		CSolverWrapper<long long, CMixedIndexPolynomialSolver<long long>>,
		// The sequence is a simple sequence of prime numbers.
		CSolverWrapper<long long, CSimplePrimeSolver<long long>>
	> solvers;

	std::getline(std::cin, line);
	done = line.empty();
	while (!done) {
		std::size_t end = 0;
		std::size_t start = 0; 
		long long number = 0;

		while (end != std::string::npos) {
			number = std::strtoll(line.data() + start, 0, 0);
			SolverInvoker<decltype(solvers)>::addNumber<decltype(number)>(solvers, number);
			end = line.find_first_of(',', start);
			if (end != std::string::npos)
				start = end + 1;
		}

		decltype(number) numbers[8];
		std::size_t numberCount = sizeof(numbers) / sizeof(decltype(number));

		SolverInvoker<decltype(solvers)>::prepare(solvers);
		if (SolverInvoker<decltype(solvers)>::compute<decltype(number)>(solvers, numbers, numberCount)) {
			for (std::size_t i = 0; i < numberCount - 1; ++i)
				std::cout << numbers[i] << ",";

			std::cout << numbers[numberCount - 1] << std::endl;
		} else std::cout << "No matching series found" << std::endl;

		SolverInvoker<decltype(solvers)>::clear(solvers);
		done = std::cin.eof();
		std::getline(std::cin, line);
	}

	return 0;
}
