
#ifndef __GAUSS_H__
#define __GAUSS_H__


#include <vector>


/// <summary>
/// Solves series of linear equations which is useful for finding
/// the right coefficients for various formulas.
/// </summary>
 template<std::size_t Rows, typename NumberType>
class CGaussElimination {
public:
	CGaussElimination() :
		rowsFilled_(0), computed_(false) { }
	/// <summary>
	/// Initializes one row of the matrix. The row must have
	/// ROws+1 items (the resulting matrix is a square, plus the right side).
	/// </summary>
	/// <param name="aRow"></param>
	/// <returns>Returns a reference to the this instance of CGauusEquation,
	/// so you can chain the addRow calls in one statement. </returns>
	CGaussElimination<Rows, NumberType> & addRow(const std::initializer_list<NumberType> & aRow)
	{
		std::size_t index = 0;

		if (aRow.size() != Rows + 1)
			throw std::exception("CGaussElimination: Invalid row size");

		if (rowsFilled_ >= Rows)
			throw std::exception("Attempt to insert a row to a fully initialized matrix");

		for (auto n : aRow) {
			matrix_[rowsFilled_][index] = n;
			++index;
		}

		++rowsFilled_;

		return *this;
	}
	/// <summary>
	/// Initializes one row of the matrix. The row must have
	/// ROws+1 items (the resulting matrix is a square, plus the right side).
	/// </summary>
	/// <param name="aRow"></param>
	/// <returns>Returns a reference to the this instance of CGauusEquation,
	/// so you can chain the addRow calls in one statement. </returns>
	CGaussElimination<Rows, NumberType> & addRow(const std::vector<NumberType> & aRow)
	{
		std::size_t index = 0;

		if (aRow.size() != Rows + 1)
			throw std::exception("CGaussElimination: Invalid row size");

		if (rowsFilled_ >= Rows)
			throw std::exception("Attempt to insert a row to a fully initialized matrix");

		for (auto n : aRow) {
			matrix_[rowsFilled_][index] = n;
			++index;
		}

		++rowsFilled_;

		return *this;
	}
	/// <summary>
	/// The Gaussian elimination over the inner matrix. The matrix needs to be
	/// fully initialized.
	/// </summary>
	void compute()
	{
		if (rowsFilled_ < Rows)
			throw std::exception("The matrix is not fully initialized");

		for (std::size_t c = 1; c < Rows; ++c) {
			bool nonZero = get(c - 1, c - 1) != 0;
			
			if (!nonZero) {
				for (std::size_t i = c; i < Rows; ++i) {
					if (get(i, c - 1) != 0) {
						swapRows(c - 1, i);
						nonZero = true;
						break;
					}
				}
			}
			
			if (nonZero) {
				for (std::size_t i = c; i < Rows; ++i) {
					NumberType multiplier = get(i, c - 1) / get(c - 1, c - 1);

					for (size_t k = c - 1; k <= Rows; ++k)
						set(i, k, get(i, k) - multiplier*get(c - 1, k));
				}
			}
		}

		for (std::size_t i = Rows; i > 0; --i) {
			NumberType sum = 0;

			for (std::size_t j = Rows; j > i; --j)
				sum += get(i - 1, j - 1)*results_[j - 1];

			if (get(i - 1, i - 1) == 0)
				results_[i - 1] = 0;
			else results_[i - 1] = (get(i - 1, Rows) - sum) / get(i - 1, i - 1);
		}

		computed_ = true;

		return;
	}
	/// <summary>
	/// Indicates whether the Gaussian elimination had been performed
	/// on this instance.
	/// </summary>
	/// <returns>Returns a boolean value idnicating whether the G. E. had finished.</returns>
	bool computed() const { return computed_; }
	/// <summary>
	/// Retrieves nth solution of the linear equations represented
	/// by the inner matrix.
	/// </summary>
	/// <param name="aIndex">The column index of the solution to retrieve. </param>
	/// <returns>Returns the nth solution.</returns>
	NumberType result(std::size_t aIndex) const 
	{
		if (!computed_)
			throw std::exception("The results have not yet been computed");

		return results_[aIndex];
	}
private:
	/// <summary>
	/// The inner matrix.
	/// </summary>
	NumberType matrix_[Rows][Rows + 1];
	/// <summary>
	/// Results of the Gaussian elimination.
	/// </summary>
	NumberType results_[Rows];
	/// <summary>
	/// Indicates whether the Gaussian elimination had been performed.
	/// </summary>
	bool computed_;
	/// <summary>
	/// The number of rows of the inner matrix that are currently initialized.
	/// The Gaussian elimination can be performed only when all the rows are
	/// initialized.
	/// </summary>
	std::size_t rowsFilled_;
	/// <summary>
	/// Swaps two rows of the inner matrix, given their row indices.
	/// </summary>
	/// <param name="aRow1">The row index of the first row.</param>
	/// <param name="aRow2">The row index of the second row.</param>
	void swapRows(std::size_t aRow1, std::size_t aRow2)
	{
		for (std::size_t i = 0; i <= Rows; ++i) {
			NumberType tmp = matrix_[aRow1][i];
			matrix_[aRow1][i] = matrix_[aRow2][i];
			matrix_[aRow2][i] = tmp;
		}

		return;
	}
	/// <summary>
	/// Retrieves an element of the inner matrix, given its position.
	/// </summary>
	/// <param name="aRow">The row index</param>
	/// <param name="aCol">The column index</param>
	/// <returns>Returns the element. </returns>
	NumberType get(std::size_t aRow, std::size_t aCol) const { return matrix_[aRow][aCol]; }
	/// <summary>
	/// Changes an element of the inner matrix.
	/// </summary>
	/// <param name="aRow">The row index. </param>
	/// <param name="aCol">The column index.</param>
	/// <param name="aValue">The new value.</param>
	void set(std::size_t aRow, std::size_t aCol, NumberType aValue) { matrix_[aRow][aCol] = aValue; }
};





#endif
