
#ifndef __SOLVER_H__
#define __SOLVER_H__


#include <cstdlib>
#include <vector>
#include "gauss.h"


/// <summary>
/// Exposes a solver instance to the public world.
/// 
/// Each solver must publish the following attributes and methods:
/// 1) std::size_t DependenciesCount
///    Number of elements needed to compute the next element in the sequence.
/// 2) void prepare(const NumberType aNumbers[], std::size_t aCount)
///    The solver gets initial part of the sequence and should initialize itself in order
///    to predict the sequence further.
/// 3) NumberType formula(std::size_t aIndex, NumberType aInputs[], std::size_t aCount)
///    Computes an element of the sequence, given its index and DependenciesCount previous
///    elements.
/// 
/// Since we are using templates rather than classic virtual methods, the signatures
/// need not to be exactly like specified here, however, must be compatible.
/// </summary>
template <typename NumberType, typename SolverType>
class CSolverWrapper {
public:
	CSolverWrapper() : numberOfDependencies_(solver_.DependenciesCount) { }
	/// <summary>
	/// Checks whether the (inner) solvers instance is able to produce
	/// additional numbers belonging to the sequnce introduced via calls
	/// to addNumber method.
	/// </summary>
	/// <returns>Returns true if the solver is capable of producing additional
	/// numbers belonging to the sequence. Returns false otherwise.</returns>
	bool check()
	{
		bool ret = true;
		NumberType res;
		std::vector<NumberType> inputs;

		for (std::size_t i = numberOfDependencies_; i < numbers_.size(); ++i) {
			for (std::size_t j = 0; j < numberOfDependencies_; ++j)
				inputs.push_back(numbers_[i - numberOfDependencies_ + j]);

			res = solver_.formula(i, inputs.data(), numberOfDependencies_);
			ret = (ret && res == numbers_[i]);
			inputs.clear();
		}

		return ret;
	}
	/// <summary>
	/// Examines the numbers given via the addNumber method and attempts to find
	/// a formula describing the sequence. The result of the operation needs to be checked
	/// by a call to the check method.
	/// </summary>
	void prepare(void)
	{
		solver_.prepare(numbers_.data(), numbers_.size());
	}
	/// <summary>
	/// Computes additional members of the sequence.
	/// </summary>
	/// <param name="aCount">Number of elements to compute.</param>
	void computeNext(std::size_t aCount)
	{
		NumberType *inputs = new NumberType[numberOfDependencies_];

		for (std::size_t i = 0; i < aCount; ++i) {
			for (std::size_t j = 0; j < numberOfDependencies_; ++j)
				inputs[j] = numbers_[numbers_.size() - numberOfDependencies_ + j];

			addNumber(solver_.formula(numbers_.size(), inputs, numberOfDependencies_));
		}

		delete[] inputs;

		return;
	}
	/// <summary>
	/// Adds an element to the sequence.
	/// </summary>
	/// <param name="aNumber">The value to add.</param>
	void addNumber(NumberType aNumber) { numbers_.push_back(aNumber); }
	/// <summary>
	///  Retrieves nth element of a sequence.
	/// </summary>
	/// <param name="aIndex">The index of the element to retrieve.</param>
	/// <returns>the aIndex-th element of the sequence.</returns>
	NumberType getNumber(std::size_t aIndex) const { return numbers_[aIndex]; }
	/// <summary>
	/// Retrieves the number of elements in the sequence.
	/// </summary>
	/// <returns>the number of elements in the sequence.</returns>
	std::size_t size() const { return numbers_.size(); }
	/// <summary>
	/// Prepares the solver instance for receiving a brand new sequence.
	/// </summary>
	void clear() { numbers_.clear(); }
protected:
	/// <summary>
	/// CUrrently known sequence elements (either computed or added via the addNumber
	/// method).
	/// </summary>
	std::vector<NumberType> numbers_;
	/// <summary>
	/// Number of (the last) sequence elements required to compute a new sequence element.
	/// </summary>
	std::size_t numberOfDependencies_;
	/// <summary>
	/// The solver instance.
	/// </summary>
	SolverType solver_;
};

template <typename NumberType, std::size_t NumberOfDependencies>
class CLinearSolver {
public:
	static const std::size_t DependenciesCount = NumberOfDependencies;
	void prepare(const NumberType aNumbers[], std::size_t aCount)
	{
		std::vector<double> oneRow;
		CGaussElimination<NumberOfDependencies + 1, double> ge;

		for (std::size_t i = 0; i < NumberOfDependencies + 1; ++i) {
			oneRow.push_back(1);
			for (std::size_t j = 0; j < NumberOfDependencies + 1; ++j)
				oneRow.push_back((double)aNumbers[i + j]);

			ge.addRow(oneRow);
			oneRow.clear();
		}

		ge.compute();
		for (std::size_t i = 0; i < NumberOfDependencies + 1; ++i)
			coefficients_[i] = ge.result(i);

		return;
	}
	NumberType formula(std::size_t aIndex, NumberType aInputs[], std::size_t aCount)
	{
		double res = coefficients_[0];

		for (std::size_t i = 0; i < NumberOfDependencies; ++i)
			res += coefficients_[i + 1] * aInputs[i];

		return (NumberType)res;
	}
private:
	double coefficients_[NumberOfDependencies + 1];
};

template <typename NumberType, std::size_t MaxDegree>
class CPolynomialSolver {
public:
	static const std::size_t DependenciesCount = 1;
	void prepare(const NumberType aNumbers[], std::size_t aCount)
	{
		std::vector<double> oneRow;
		CGaussElimination<MaxDegree + 1, double> ge;

		for (std::size_t i = 0; i < MaxDegree + 1; ++i) {
			double item = 1;

			for (std::size_t j = 0; j < MaxDegree + 1; ++j) {
				oneRow.push_back(item);
				item *= aNumbers[i];
			}

			oneRow.push_back((double)aNumbers[i + 1]);
			ge.addRow(oneRow);
			oneRow.clear();
		}

		ge.compute();
		for (std::size_t i = 0; i < MaxDegree + 1; ++i)
			coefficients_[i] = ge.result(i);

		return;
	}
	NumberType formula(std::size_t aIndex, NumberType aInputs[], std::size_t aCount)
	{
		NumberType x = aInputs[0];
		double res = coefficients_[0];

		for (std::size_t i = 0; i < MaxDegree; ++i) {
			res += coefficients_[i + 1] * x;
			x *= aInputs[0];
		}

		return (NumberType)res;
	}
private:
	double coefficients_[MaxDegree + 1];
};

template <typename NumberType, std::size_t MaxDegree>
class CIndexSolver {
public:
	static const std::size_t DependenciesCount = 1;
	void prepare(const NumberType aNumbers[], std::size_t aCount)
	{
		bool ret = true;
		std::vector<double> oneRow;
		CGaussElimination<MaxDegree + 1, double> ge;

		for (std::size_t i = 0; i < MaxDegree + 1; ++i) {
			double item = 1;

			for (std::size_t j = 0; j < MaxDegree + 1; ++j) {
				oneRow.push_back(item);
				item *= i;
			}

			oneRow.push_back((double)aNumbers[i]);
			ge.addRow(oneRow);
			oneRow.clear();
		}

		ge.compute();
		for (std::size_t i = 0; i < MaxDegree + 1; ++i)
			coefficients_[i] = ge.result(i);

		return;
	}
	NumberType formula(std::size_t aIndex, NumberType aInputs[], std::size_t aCount)
	{
		NumberType x = aIndex;
		double res = coefficients_[0];

		for (std::size_t i = 0; i < MaxDegree; ++i) {
			res += coefficients_[i + 1] * x;
			x *= aIndex;
		}

		return (NumberType)res;
	}
private:
	double coefficients_[MaxDegree + 1];
};


template <typename NumberType>
class CMixedIndexPolynomialSolver {
public:
	static const std::size_t DependenciesCount = 1;
	void prepare(const NumberType aNumbers[], std::size_t aCount)
	{
		bool ret = true;
		std::vector<double> oneRow;
		CGaussElimination<7, double> ge;

		for (std::size_t i = 0; i < 7; ++i) {
			const double ai = (double)aNumbers[i];
			const double aip1 = (double)aNumbers[i + 1];
			const std::size_t index = i + 1;

			oneRow.push_back(1);
			oneRow.push_back(index);
			oneRow.push_back(index*index);
			oneRow.push_back(ai);
			oneRow.push_back(ai*ai);
			oneRow.push_back(index*ai);
			oneRow.push_back(index*index*ai*ai);

			oneRow.push_back(aip1);
			ge.addRow(oneRow);
			oneRow.clear();
		}

		ge.compute();
		for (std::size_t i = 0; i < 7; ++i)
			coefficients_[i] = ge.result(i);

		return;
	}
	NumberType formula(std::size_t aIndex, NumberType aInputs[], std::size_t aCount)
	{

		return (NumberType)(
			coefficients_[0] +
			coefficients_[1]*aIndex +
			coefficients_[2]*aIndex*aIndex +
			coefficients_[3]*aInputs[0] +
			coefficients_[4]*aInputs[0]*aInputs[0] +
			coefficients_[5]*aInputs[0]*aIndex +
			coefficients_[6] * aIndex*aIndex*aInputs[0] * aInputs[0]
			);
	}
private:
	double coefficients_[7];
};

template <typename NumberType>
class CSimplePrimeSolver {
public:
	CSimplePrimeSolver() 
		: allPrimes_(false) {}
	static const std::size_t DependenciesCount = 1;
	void prepare(const NumberType aNumbers[], std::size_t aCount)
	{
		allPrimes_ = true;

		for (std::size_t i = 0; i < aCount; ++i) {
			allPrimes_ = isPrime(aNumbers[i]);
			if (!allPrimes_)
				break;
		}

		return;
	}
	NumberType formula(std::size_t aIndex, NumberType aInputs[], std::size_t aCount)
	{
		NumberType ret = 0;

		if (allPrimes_) {
			ret = aInputs[0];
			if (ret == 1 || ret == 2)
				ret += 1;
			else {
				ret = aInputs[0] + 2;
				while (!isPrime(ret))
					ret += 2;
			}
		}

		return ret;
	}
private:
	bool allPrimes_;
	bool isPrime(NumberType aValue)
	{
		bool ret = (aValue == 1 || aValue == 2);
		std::size_t maximum = (std::size_t)round(sqrt(aValue));

		if (aValue >= 3) {
			ret = ((aValue % 2) != 0);
			if (ret) {
				for (std::size_t i = 3; i <= maximum; i += 2) {
					ret = (aValue % i) != 0;
					if (!ret)
						break;
				}
			}
		}

		return ret;
	}
};



#endif
